/*
    Copyright 2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/
// win:   gcc -o myproxocket.dll myproxocket.c -shared
// linux: gcc -o myproxocket.so myproxocket.c -shared -ldl
//        export LD_PRELOAD=./myproxocket.so

#ifndef WIN32
    #define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>
#include <sys/stat.h>

#ifdef WIN32
    #include <winsock.h>
    #include <windows.h>
    typedef int socklen_t;
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <dlfcn.h>  // -ldl
    #define myrecvfrom recvfrom
    static ssize_t (*old_recvfrom)(int s, void *buf, size_t len, int flags, struct sockaddr *from, socklen_t *fromlen) = NULL;
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER             "0.1.1"
#define PACKET_FNAME    "packet.dat"    // name of the file containing the beginning of the packet to filter
#define MIN_SECONDS     15              // seconds between each player from the same IP
#define MAXIPS          256             // enough, it MUST be related on MIN_SECONDS
                                        // so if MAXIPS is 256 and MIN_SECONDS is 15 means that
                                        // this plugin can support about a new player from different IP
                                        // addresses each 58 milliseconds... definitely more than enough

typedef struct {
    u32     ip;     // IP address
    time_t  last;   // last time that the packet has been received
} ips_t;

static ips_t    ips[MAXIPS] = {{0,0}};  // init
static u8       *packet     = NULL;     // the beginning of the packet to monitor
static int      packet_size = 0;        // the amount of bytes to match the packet



u8 *fdload(u8 *fname, int *fsize) {     // loads packet.dat in memory
    struct stat xstat;
    FILE    *fd;
    int     size;
    u8      *buff;

    fd = fopen(fname, "rb");
    if(!fd) exit(1);
    fstat(fileno(fd), &xstat);
    size = xstat.st_size;
    buff = malloc(size);
    if(!buff) exit(1);
    fread(buff, 1, size, fd);
    fclose(fd);
    *fsize = size;
    return(buff);
}



ssize_t myrecvfrom(int s, void *buf, size_t len, int flags, struct sockaddr *from, socklen_t *fromlen) {
    time_t  now;
    u32     ip;
    int     i,
            older;

#ifndef WIN32
    if(!old_recvfrom) old_recvfrom = dlsym(RTLD_NEXT, "recvfrom");
    len = old_recvfrom(s, buf, len, flags, from, fromlen);
#endif

    if(!from) return(len);
    if((int)len <= 0) return(len);

    if(!packet) {   // load the packet, it's also possible to place this instruction in the init
        packet = fdload(PACKET_FNAME, &packet_size);
    }

    if(((int)len >= packet_size) && !memcmp(buf, packet, packet_size)) {
        ip   = ((struct sockaddr_in *)from)->sin_addr.s_addr;
        now  = time(NULL);

        for(i = 0; i < MAXIPS; i++) {   // check if the IP is already in the database
            if(ip != ips[i].ip) continue;
            if(now < (ips[i].last + MIN_SECONDS)) {
                len = 0;                // set the length of the packet to zero so he can't join
            }
            ips[i].last = now;          // update the timer
            return(len);
        }

        older = 0;
        for(i = 0; i < MAXIPS; i++) {   // add or update the IP
            if(!ips[i].ip) {            // empty place, not so useful
                older = i;              // because it's related to MAXIPS
                break;
            }
            if(ips[i].last < ips[older].last) {
                older = i;              // find the oldest IP
            }
        }
        ips[older].ip   = ip;
        ips[older].last = now;
    }
    return(len);
}



#ifdef WIN32
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpReserved) {
    switch(fdwReason) {
        case DLL_PROCESS_ATTACH: {
            DisableThreadLibraryCalls(hinstDLL);
            break;
        }
        case DLL_PROCESS_DETACH: {
            break;
        }
        default: break;
    }
    return(TRUE);
}
#endif

